'''
Module with tests for BkkChecker class
'''
from importlib.resources import files

import pytest

from dmu.logging.log_store                import LogStore
from ap_utilities.bookkeeping.bkk_checker import BkkChecker

# ----------------------------------------
@pytest.fixture(scope='session', autouse=True)
def _initialize():
    LogStore.set_level('ap_utilities:Bookkeeping.bkk_checker', 10)
# ----------------------------------------
def test_simple():
    '''
    Will save list of samples to YAML
    '''
    samples_path = files('ap_utilities_data').joinpath('rd_samples.yaml')
    samples_path = str(samples_path)

    obj=BkkChecker(samples_path)
    obj.save()
# ----------------------------------------
def test_multithreaded():
    '''
    Will save list of samples to YAML using 4 threads
    '''
    samples_path = files('ap_utilities_data').joinpath('rd_samples.yaml')
    samples_path = str(samples_path)

    obj=BkkChecker(samples_path)
    obj.save(nthreads=8)
# ----------------------------------------



'''
Script used to validate ntuples produced by AP pipelines
'''
import os
import glob
import shutil
import argparse
from typing              import Union
from typing              import ClassVar
from importlib.resources import files
from dataclasses         import dataclass

import tqdm
import yaml
from ROOT                  import TFile, TDirectoryFile, TTree
from dmu.logging.log_store import LogStore

log = LogStore.add_logger('ap_utilities_scripts:validate_ap_tuples')
# -------------------------------
@dataclass
class Data:
    '''
    Class holding shared attributes
    '''
    pipeline_id : int
    config_name : str
    cfg         : dict

    d_tree_miss : ClassVar[dict[str, list[str]]] = {}
    d_tree_found: ClassVar[dict[str, list[str]]] = {}
    d_tree_entries : ClassVar[dict[str, dict[str,int]]] = {}
# -------------------------------
def _check_path(path : str) -> None:
    found = os.path.isdir(path) or os.path.isfile(path)
    if not found:
        raise FileNotFoundError(f'Cannot find: {path}')
# -------------------------------
def _parse_args() -> None:
    parser = argparse.ArgumentParser(description='Makes a list of PFNs for a specific set of eventIDs in case we need to reprocess them')
    parser.add_argument('-p','--pipeline', type=int, help='Pipeline ID', required=True)
    parser.add_argument('-c','--config'  , type=str, help='Name of config file, without extension', required=True)
    parser.add_argument('-l','--log_lvl' , type=int, help='Logging level', default=20, choices=[10,20,30])
    args = parser.parse_args()

    Data.pipeline_id = args.pipeline
    Data.config_name = args.config

    LogStore.set_level('ap_utilities_scripts:validate_ap_tuples', args.log_lvl)
# -------------------------------
def _load_config() -> None:
    config_path = files('ap_utilities_data').joinpath(f'{Data.config_name}.yaml')
    config_path = str(config_path)

    if not os.path.isfile(config_path):
        raise FileNotFoundError(f'Could not find: {config_path}')

    with open(config_path, encoding='utf-8') as ifile:
        Data.cfg = yaml.safe_load(ifile)
# -------------------------------
def _get_out_paths() -> list[str]:
    '''
    Returns list of paths containing the ROOT files and the zip file with the logs
    '''
    pipeline_dir = Data.cfg['paths']['pipeline_dir']
    analysis_dir = Data.cfg['paths']['analysis_dir']

    job_path = f'{pipeline_dir}/{Data.pipeline_id}/{analysis_dir}'
    _check_path(job_path)

    sample_wc = f'{job_path}/*/*'
    l_sample  = glob.glob(sample_wc)

    nsample   = len(l_sample)
    njobs     = len(Data.cfg['samples'])

    #if nsample != njobs:
    #    raise ValueError(f'Number of samples and jobs in {sample_wc} differ: {nsample} -> {njobs}')

    return l_sample
# -------------------------------
def _get_file_path(job_path : str, ending : str) -> Union[str,None]:
    path_wc = f'{job_path}/*{ending}'
    try:
        [file_path] = glob.glob(path_wc)
    except ValueError:
        log.warning(f'Cannot find one and only one file in: {path_wc}')
        return None

    return file_path
# -------------------------------
def _sample_from_root(root_path : str) -> str:
    l_samp = Data.cfg['samples']
    try:
        [samp] = [ sample for sample in l_samp if sample in root_path ]
    except ValueError as exc:
        raise ValueError(f'Not found one and only one sample corresponding to: {root_path}') from exc

    return samp
# -------------------------------
def _copy_path(source : str) -> str:
    target = f'/tmp/{source}'
    target = target.replace('#', '_')

    if os.path.isfile(target):
        return target

    target_dir = os.path.dirname(target)
    os.makedirs(target_dir, exist_ok=True)

    log.info(f'{source} --> {target}')
    shutil.copy(source, target)

    return target
# -------------------------------
def _lines_from_samples(l_samp : list[str]) -> set[str]:
    l_line = []
    for samp in l_samp:
        l_line += Data.cfg['analyses'][samp]

    return set(l_line)
# -------------------------------
def _validate_root_file(root_path : str) -> None:
    _validate_trees(root_path)
# -------------------------------
def _is_valid_dir(sample : str, file_dir : TDirectoryFile) -> bool:
    if not hasattr(file_dir, 'DecayTree') or not isinstance(file_dir.DecayTree, TTree):
        Data.d_tree_entries[sample] = {file_dir.GetName : 0}
        return False

    nentries = file_dir.DecayTree.GetEntries()
    if nentries == 0:
        Data.d_tree_entries[sample] = {file_dir.GetName : 0}
        return False

    Data.d_tree_entries[sample] = {file_dir.GetName() : nentries}

    return True
# -------------------------------
def _validate_trees(root_path : str) -> None:
    sample    = _sample_from_root(root_path)
    l_samp    = Data.cfg['samples'][sample]
    s_line    = _lines_from_samples(l_samp)

    root_path = _copy_path(root_path)
    rfile     = TFile(root_path)
    l_key     = rfile.GetListOfKeys()
    l_dir     = [ key.ReadObj() for key in l_key if key.ReadObj().InheritsFrom('TDirectoryFile') ]
    s_name    = { fdir.GetName() for fdir in l_dir if _is_valid_dir(sample, fdir)}

    s_missing = s_line - s_name
    if s_name != s_line:
        log.debug(f'File: {root_path}')
        log.debug(f'Missing : {s_missing}')

    ntree     = len(l_dir)
    log.debug(f'Found {ntree} trees for sample {sample}')

    if len(s_missing) != 0:
        _save_trees(sample, s_missing, Data.d_tree_miss )

    _save_trees(sample,    s_name, Data.d_tree_found)
# -------------------------------
def _save_trees(sample : str, s_tree_name : set[str], d_data : dict[str, list[str]]):
    l_tree_name = list(s_tree_name)
    d_data.update({sample : l_tree_name})
# -------------------------------
def _validate_job(job_path : str) -> None:
    '''
    Picks path to directory with ROOT and zip file
    Runs validation
    '''
    root_path = _get_file_path(job_path, ending='_2.tuple.root')
    log_path  = _get_file_path(job_path, ending=         '.zip')

    if log_path is None or root_path is None:
        return

    _validate_root_file(root_path)
# -------------------------------
def _validate() -> None:
    _load_config()
    l_out_path = _get_out_paths()

    for out_path in tqdm.tqdm(l_out_path, ascii=' -'):
        _validate_job(out_path)
# -------------------------------
def _save_report() -> None:
    d_rep = {
            'missing_trees' : Data.d_tree_miss,
            'found_trees'   : Data.d_tree_found,
            'tree_entries'  : Data.d_tree_entries,
            }

    with open('report.yaml', 'w', encoding='utf-8') as ofile:
        yaml.safe_dump(d_rep, ofile)
# -------------------------------
def main():
    '''
    Script starts here
    '''
    _parse_args()
    _validate()
    _save_report()
# -------------------------------
if __name__ == '__main__':
    main()



'''
Module with BkkChecker class
'''

import re
from concurrent.futures     import ThreadPoolExecutor

import subprocess
import yaml

import dmu.physics.utilities as phut
from dmu.logging.log_store  import LogStore

log=LogStore.add_logger('ap_utilities:Bookkeeping.bkk_checker')
# ---------------------------------
class BkkChecker:
    '''
    Class meant to check if samples exist in Bookkeeping using multithreading.
    This is useful with large lists of samples, due to low performance of Dirac
    '''
    # pylint: disable=too-few-public-methods
    # -------------------------
    def __init__(self, path : str):
        '''
        Takes the path to a YAML file with the list of samples
        '''
        with open(path, encoding='utf-8') as ifile:
            self._d_cfg                    = yaml.safe_load(ifile)
            self._l_event_type : list[str] = self._d_cfg['event_type']

        self._input_path   : str = path
        self._year         : str = self._d_cfg['settings']['year']
        self._mc_path      : str = self._d_cfg['settings']['mc_path']
        self._nu_path      : str = self._d_cfg['settings']['nu_path']
        self._polarity     : str = self._d_cfg['settings']['polarity']
        self._generator    : str = self._d_cfg['settings']['generator']
        self._sim_version  : str = self._d_cfg['settings']['sim_vers']
        self._ctags        : str = self._d_cfg['settings']['ctags']
        self._dtags        : str = self._d_cfg['settings']['dtags']
    # -------------------------
    def _nfiles_line_from_stdout(self, stdout : str) -> str:
        l_line = stdout.split('\n')
        try:
            [line] = [ line for line in l_line if line.startswith('Nb of Files') ]
        except ValueError:
            log.warning(f'Cannot find number of files in: \n{stdout}')
            return 'None'

        return line
    # -------------------------
    def _nfiles_from_stdout(self, stdout : str) -> int:
        line  = self._nfiles_line_from_stdout(stdout)
        log.debug(f'Searching in line {line}')

        regex = r'Nb of Files      :  (\d+|None)'
        mtch  = re.match(regex, line)

        if not mtch:
            raise ValueError(f'No match found in: \n{stdout}')

        nsample = mtch.group(1)
        if nsample == 'None':
            log.debug('Found zero files')
            return 0

        log.debug(f'Found {nsample} files')

        return int(nsample)
    # -------------------------
    def _was_found(self, event_type : str) -> bool:
        sample_path = f'/MC/{self._year}/Beam6800GeV-{self._mc_path}-{self._polarity}-{self._nu_path}-25ns-{self._generator}/{self._sim_version}/HLT2-{self._mc_path}/{event_type}/DST'

        log.debug(f'{"":<4}{sample_path:<100}')

        cmd_bkk = ['dirac-bookkeeping-get-stats', '-B' , sample_path]
        result  = subprocess.run(cmd_bkk, capture_output=True, text=True, check=False)
        nfile   = self._nfiles_from_stdout(result.stdout)

        return nfile != 0
    # -------------------------
    def _get_samples_with_threads(self, nthreads : int) -> list[str]:
        l_found : list[bool] = []
        with ThreadPoolExecutor(max_workers=nthreads) as executor:
            l_result = [ executor.submit(self._was_found, event_type) for event_type in self._l_event_type ]
            l_found  = [result.result() for result in l_result ]

        l_event_type = [ event_type for event_type, found in zip(self._l_event_type, l_found) if found ]

        return l_event_type
    # -------------------------
    def _save_to_text(self, l_event_type : list[str]) -> None:
        text = ''
        for evt_type in l_event_type:
            nu_name   = self._nu_path.replace('.', 'p')
            nick_name = phut.read_decay_name(evt_type)
            text   += f'("{nick_name}", "{evt_type}" , "{self._mc_path}", "{self._polarity}"  , "{self._ctags}", "{self._dtags}", "{self._nu_path}", "{nu_name}", "{self._sim_version}", "{self._generator}" ),\n'

        output_path = self._input_path.replace('.yaml', '.txt')

        log.info(f'Saving to: {output_path}')
        with open(output_path, 'w', encoding='utf-8') as ofile:
            ofile.write(text)
    # -------------------------
    def save(self, nthreads : int = 1) -> None:
        '''
        Will check if samples exist in grid
        Will save list of found samples to text file with same name as input YAML, but with txt extension
        '''

        log.info('Filtering input')
        if nthreads == 1:
            log.info('Using single thread')
            l_event_type = [ event_type for event_type in self._l_event_type if self._was_found(event_type) ]
        else:
            log.info(f'Using {nthreads} threads')
            l_event_type = self._get_samples_with_threads(nthreads)

        nfound = len(l_event_type)
        npased = len(self._l_event_type)

        log.info(f'Found: {nfound}/{npased}')
        self._save_to_text(l_event_type)
# ---------------------------------
